/*

This file is part of aiorOS

aiorOS - 32bit Operating System.
Copyright (C) 2005  Salvatore Iengo <salvatoreiengo@hotmail.com>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include<mem.h>
#include<task.h>
#include<fs.h>
#include<swapper.h>
#include<idt.h>

/*
 * BSS data
 */
PageDirEntry *kPageDir = (PageDirEntry*)KERNEL_CACHE;
PageTableEntry *kPageTable = (PageTableEntry*)(KERNEL_CACHE + PAGE_SIZE);
PageTableEntry *kPageShared = (PageTableEntry*)(KERNEL_CACHE + PAGE_SIZE*2);

static byte volatile pagesBitmap[PAGES_BITMAP_SIZE];

static int code_page_request(dword address);
static int stack_page_request(dword address);
static int table_page_present(dword dir);
static void create_table_request(void *page, int type, int mode);
static void bind_page(dword address, void *page, int mode);
static int load_page(char *path, void *buffer, int pageNo);
static void initPagesBitmap(void);

void initMem(void){
   initPagesBitmap();

   kPageTable = create_page_table(kPageTable, PAGE_TYPE_KERNEL, PAGE_MODE_USER);
   kPageDir = create_page_dir(kPageDir, kPageTable);
   kPageShared = create_page_table(kPageShared, PAGE_TYPE_EMPTY, PAGE_MODE_SYSTEM);
   kPageDir[1023]=PAGE_DIR_ENTRY(PAGE_ENTRY_PRESENT,PAGE_MODE_SYSTEM,(((dword)kPageShared)>>12));

   loadPageDirRegister(kPageDir);
   enablePaging();
}

void callback_pages_for(int pid){
}

void validate_mem(PageDirEntry shared){
}

void share_task_struct(int pid, void *address){
   kPageShared[1023-pid] = PAGE_TABLE_ENTRY(PAGE_ENTRY_PRESENT, PAGE_MODE_SYSTEM, ((dword)address)>>12);
}

void *getPages(int num){
   int i,j,k,free,last;

   for(i=0,free=0,last=-1;i<PAGES_BITMAP_SIZE;i++){
      for(j=0;j<8;j++){
         if(!IS_SET_BIT(pagesBitmap[i],j)){
            free++;
         }
         else
         {
            free=0;
            last = i*8+j;
         }
         if(free>=num){
            int byteOff, bitOff;
            last++;
            for(k=last;k<last+num;k++){
               /*set k-th bit of last...last+num interval*/
               byteOff = k / 8;
               bitOff = k % 8;
               SET_BIT(pagesBitmap[byteOff],bitOff);
            }
            return (void*)(KERNEL_MEM_LIMIT+(last*PAGE_SIZE));
         }
      }
   }
   return NULL;
}

void freePages(void *base, int num){
   int k,last;
   int byteOff,bitOff;

   if((base=(((char*)base)-KERNEL_MEM_LIMIT))<0)
      return;

   last = ((unsigned int)(base))>>12;

   for(k=last;k<last+num;k++){
      /*set k-th bit of last...last+num interval*/
      byteOff = k / 8;
      bitOff = k % 8;
      CLEAR_BIT(pagesBitmap[byteOff],bitOff);
   }
}

/*
 * Move to physical addressing
 */
dword BEGIN_PHYS_ADDR(int dis_int)
{
   dword oesp;

   if(dis_int)
      CLI();

   __asm__ volatile("\
          mov %%esp, %0\n\
          ":"=r"(oesp));

   __asm__ volatile("\
          movl $0x100000, %%eax;\n\
          cmp %%esp,%%eax;\n\
          ja ebpa;\n\
          movl %%esp, %%eax;\n\
          andl $0x00000FFF, %%eax;\n\
          addl %0, %%eax;\n\
          movl %%eax, %%esp;\n\
         ebpa:\n\
          movl %%cr0,%%eax;\n\
          andl $0x7FFFFFFF,%%eax;\n\
          movl %%eax,%%cr0;\n\
          ":
           :"r"(((dword)GET_CURRENT_TASK()->Info.Address.physical)&0xFFFFF000)
           :"%eax");

   return oesp;
}

/*
 * Move to paged addressing
 */
void END_PHYS_ADDR(dword iesp, int en_int){
   __asm__ volatile("\
          movl $0x100000, %%eax;\n\
          cmp %%esp,%%eax;\n\
          ja eepa;\n\
          movl %%esp, %%eax;\n\
          andl $0x00000FFF, %%eax;\n\
          addl %0, %%eax;\n\
          movl %%eax, %%esp;\n\
         eepa:\n\
          movl %%cr0,%%eax;\n\
          orl $0x80000000,%%eax;\n\
          movl %%eax,%%cr0;\n\
          sti;\n\
          ":
           :"r"(iesp&0xFFFFF000)
           :"%eax");

   if(en_int)
      STI();
}

PageDirEntry* create_page_dir(PageDirEntry *dir, PageTableEntry *table){
   int i;

   /*setup all pagedir entry (each pageDir enables 4MB)*/
   for(i=0;i<1024;i++)
      dir[i] = PAGE_DIR_ENTRY((i==0)?1:0, 1, ((dword)(table)>>12));
   dir[1023] = PAGE_DIR_ENTRY(1, 0, ((dword)kPageShared>>12));

   return dir;

}

PageTableEntry* create_page_table(PageTableEntry *table, int type, int mode){
   int i;
   /*setup all page table entries*/

/*kprintf("empty:%d, user:%d\n",IS_OPT(opt, TABLE_EMPTY),IS_OPT(opt, TABLE_MODE_USER));
*/
   for(i=0;i<1024;i++){
      if(type == PAGE_TYPE_EMPTY)
         table[i] = PAGE_TABLE_ENTRY(PAGE_ENTRY_NOT_PRESENT, mode, i);
      else
         table[i] = PAGE_TABLE_ENTRY((i<KERNEL_PAGE_SIZE)?PAGE_ENTRY_PRESENT:PAGE_ENTRY_NOT_PRESENT, (i<KERNEL_PAGE_SIZE)?PAGE_MODE_SYSTEM:mode, i);
   }

   return table;
}

#define _DEBUG_NO
void on_page_fault(dword address, dword code){
   union task_t *task = GET_CURRENT_TASK();
   PageDirEntry *pdp;
   int dir, table, offset, table_present;
   void *page;

   pdp = (PageDirEntry*)storePageDirRegister();

   dir = DIR(address);
   table = TABLE(address);
   offset = OFFSET(address);

   table_present=table_page_present(dir);

   if(dir!=1023)
   /* 1023 is a shared page */
      if(!(page = getPage()))
         swapper(1);/* invoke swapper requesting 1 page */

   #ifdef _DEBUG_
   kprintf("page fault: <%d><%d><%d>, US: %d\n",dir,table,offset,IS_USER_FAULT(code));
   #endif

   if(IS_USER_FAULT(code)){
   /* user mode fault (don't trust) */

      if(table_present){
      /* page request */

         if(code_page_request(address)){
            if(!swap_in(task->Info.pid, address, page))
               load_page(task->Info.name, page, (address-KERNEL_MEM_LIMIT)/PAGE_SIZE);
         }
         else if(stack_page_request(address)){
            swap_in(task->Info.pid, address, page);
         }
         else{
            /* raise a stack fault
            exception (#12)*/
            exceptionHandler(12, 0);
         }

      }
      else{
      /* page table request */
         if(!swap_in(task->Info.pid, address, page)){
            create_table_request(page, PAGE_TYPE_EMPTY, PAGE_MODE_USER);
         }

      }

      /* bind page to address */
      bind_page(address, page, 1);

   }
   else{
   /* system mode fault (trust) */

      if(dir == 1023){
      /* shared page */
         if(!table_present)
            bind_page(address, (void*)kPageShared,0);
      }
   }

   #ifdef _DEBUG_I_
   kscanf("%d",&i);
   #endif
}

static int code_page_request(dword address){
   union task_t *task = GET_CURRENT_TASK();

   if( ((address-KERNEL_MEM_LIMIT) < task->Info.Address.size) )
      return 1;
   else
      return 0;
}

static int stack_page_request(dword address){
   /* max 1MB of stack */
   return ((DIR(address)==1022) && (TABLE(address)>255));
}

static int table_page_present(dword dir){
   dword info, present;
   PageDirEntry *pd = (PageDirEntry*)storePageDirRegister();

   info = BEGIN_PHYS_ADDR(1);
   present = pd[dir].Present;
   END_PHYS_ADDR(info,1);

   return (int)present;
}

static void create_table_request(void *page, int type, int mode){
   dword info = BEGIN_PHYS_ADDR(1);
   create_page_table(page, type, mode);
   END_PHYS_ADDR(info,1);
}

static void bind_page(dword address, void *page, int mode){
   PageDirEntry *pdp = (PageDirEntry*)storePageDirRegister();
   PageTableEntry *tb;
   int dir, table, offset;
   dword info;

   dir = address >> 22;
   table = (address & 0x003FF000)>>12;
   offset = address & 0x00000FFF;

   info = BEGIN_PHYS_ADDR(1);

   if(pdp[dir].Present){

      if(dir==1023)
         ((union task_t *)page)->Info.Address.physical = (union task_t*)page;

      tb = (PageTableEntry*)(((dword)(pdp[dir].PageTableBaseAddress))<<12);
      tb[table] = PAGE_TABLE_ENTRY(PAGE_ENTRY_PRESENT,mode,((dword)page)>>12);
   }
   else{
      pdp[dir] = PAGE_DIR_ENTRY(PAGE_ENTRY_PRESENT, mode, ((dword)page>>12));
   }

   END_PHYS_ADDR(info,1);
}

static int load_page(char *path, void *buffer, int pageNo){
   file_t file;
   dword info;

   file = fs_open(path);

   info = BEGIN_PHYS_ADDR(1);

   fs_seek(&file, pageNo*PAGE_SIZE);
   fs_read(&file, buffer, PAGE_SIZE);
   fs_close(&file);

   END_PHYS_ADDR(info,1);

  return 1;
}


static void initPagesBitmap(void){
   int i;
   /*setup pages bitmap*/
   for(i=0;i<PAGES_BITMAP_SIZE;i++){
      pagesBitmap[i]=0; /*clearing 8 bit at time*/
   }
}

dword memFreeCount(void){
   int i,j,k,last;
   dword free;

   for(i=0,free=0;i<PAGES_BITMAP_SIZE;i++)
      for(j=0;j<8;j++)
         if(!IS_SET_BIT(pagesBitmap[i],j))
            free++;

   return free;
}
